/**************

 XPATCH2
  Example of how this program is to be used:
  
     rm -rf /tmp/.wine-*
     LD_LIBRARY_PATH=/usr/lib/wine/lib \
     LD_PRELOAD=~bisqwit/nes/xgetimage/xpatch2.so \
     /usr/lib/wine/bin/wine-pthread fam/FAMBLIP.EXE tinytoon.nes
  
  It logs audio. Requires Streamixer, LAME. (Alsa only)

****/

//#define NICE "nice -n -10 "
#define NICE ""

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>

static void LogAudio(const void*buf, unsigned n)
{
    /* Must be a local class - otherwise
     * it's reinstantiated every time
     * it loads, causing a fork bomb
     */
    static struct OutputPipe
    {
        FILE* fp;
        OutputPipe(): fp(NULL)
        {
            /* Prevent the subprocesses getting the same
             * measures as this victim program did
             */
            unsetenv("LD_PRELOAD");

            fp = popen(
                "bufferprog"
                "|"NICE" resample mb m"
                "|"NICE" lame -rxdh -m m -b 64 -q 0 - wine-audiolog.mp3",
                    "w");
        }
        ~OutputPipe()
        {
            if(fp) pclose(fp);
        }
        void Write(const void*buf, unsigned n)
        {
            fwrite(buf, 1, n, fp);
        }
    } Pipe;

    Pipe.Write(buf, n);
}

extern "C" {

int ioctl(int fd,unsigned req,void *p)
{
    /* The old function is found in libc.so.6 */
    static struct Get1
    {
        void *f;
        Get1() : f(dlopen("libc.so.6",RTLD_NOW)) {}
        ~Get1() { dlclose(f); }
        void *operator() (const char *s) { return dlsym(f, s); }
    } caller;

    /* Find the real ioctl function */
    typedef int (*old_t)(int,unsigned,void *);
    static old_t old = NULL;
    if(!old) old = (old_t)caller("ioctl");
    
    /* Check this is an ALSA function call */
    if(req == 0x400C4150  /* SNDRV_PCM_IOCTL_WRITEI_FRAMES */)
    {
        int result = old(fd, req, p);
        
        struct snd_xferi_t {
            long result;
            void *buf;
            unsigned long frames;
        }* xferi = (snd_xferi_t*) p;
        
        if(xferi->result > 0)
        {
            LogAudio(xferi->buf, xferi->result);
        }
        
        return result;
    }

    /* Call the real ioctl function */
    return old(fd,req,p);
}

int snd_pcm_hw_params_set_access(void* pcm, void* params, int access)
{
    /* The old function is found in libasound.so */
    static struct Get2
    {
        void *f;
        Get2() : f(dlopen("libasound.so", RTLD_LAZY)) {}
        ~Get2() { dlclose(f); }
        void *operator() (const char *s) { return dlsym(f, s); }
    } caller;

    /* mmap is impossible to hook, thus we sort of disable it! */
    if(access == 0 /* SND_PCM_ACCESS_MMAP_INTERLEAVED */) return -1;
    
    /* Find the real function */
    typedef int (*old_t)(void*, void*, int);
    static old_t old = NULL;
    if(!old) old = (old_t)caller("snd_pcm_hw_params_set_access");
    
    /* Call the real function */
    return old(pcm,params,access);
}

}
